package raft.karga.sound;

import raft.karga.core.*;
import raft.karga.core.Logger;
import raft.karga.world.KConfig;
import com.threed.jpct.*;

import java.util.*;
import java.io.*;
import javax.sound.sampled.*;

/**
 *
 * @author hakan eryargi (r a f t)
 */
public class PointSound {
    private final Logger logger = Logger.getLogger();
    
    private final SoundBank soundBank;
    private Map<String, ClipData> clips = new HashMap<String, ClipData>();
    
    private boolean disposed = false;
    
    private float pan = 0f;
    private float gain = 0f;
    // todo: remember last values and calculate only when necessary
    //private SimpleVector lastSourcePos, lastDestPos, lastDestDirection;
    
    /**
     * Creates a new instance of PointSound
     */
    PointSound(SoundBank soundBank) {
        this.soundBank = soundBank;
    }
    
    /** starts to play sound with calculated pan and gain values */
    public void playLoop(String group, String name, SimpleVector sourcePos, SimpleVector destPos, SimpleVector destDirection) {
        play(group, name, sourcePos, destPos, destDirection, SoundBank.PLAY_LOOP);
    }
    
    /** starts to play sound with calculated pan and gain values */
    public void playLoop(String name, SimpleVector sourcePos, SimpleVector destPos, SimpleVector destDirection) {
        play(null, name, sourcePos, destPos, destDirection, SoundBank.PLAY_LOOP);
    }
    
    /** starts to play sound with calculated pan and gain values */
    public void playOnce(String name, SimpleVector sourcePos, SimpleVector destPos, SimpleVector destDirection) {
        play(null, name, sourcePos, destPos, destDirection, SoundBank.PLAY_ONCE);
    }
    
    /** starts to play sound with calculated pan and gain values */
    private void play(String group, String name, SimpleVector sourcePos, SimpleVector destPos, SimpleVector destDirection, final boolean loop) {
        if (KConfig.mute || (name == null))
            return;
        
        checkDisposed();
        
        if (group == null) {
            if (clips.containsKey(name))
                return;
        } else {
            ClipData clipData = clips.get(group);
            if (clipData != null) {
                if (name.equals(clipData.name)) {
                    return;
                } else {
                    clipData.close();
                    clips.remove(group);
                }
            }
        }
        if (tooFarToPlay(sourcePos, destPos))
            return;
        
        try {
            final ClipData clipData = new ClipData(soundBank, name);
            
            pan = calculatePan(sourcePos, destPos, destDirection);
            gain = calculateGain(sourcePos, destPos, destDirection);
            
            setPanAndGain(clipData, pan, gain);
            /*
            if (loop)
                clipData.clip.loop(Clip.LOOP_CONTINUOUSLY);
            else clipData.clip.start();
             */
            
            if (loop)
                clipData.loop();
            else clipData.start();
            /*
            if (clipData.clip != null) {
                if (loop)
                    clipData.clip.loop(Clip.LOOP_CONTINUOUSLY);
                else clipData.clip.start();
            } else {
                new Thread() {
                    public void run() {
                        clipData.sourceDataLine.start();
                        do {
                            clipData.sourceDataLine.write(clipData.soundData.bytes, 0, clipData.soundData.bytes.length);
                            //System.out.println("wrote " + clipData.soundData.bytes.length + " bytes");
                            clipData.sourceDataLine.drain();
                            //System.out.println("drained");
                        } while (loop && clipData.sourceDataLine.isRunning());
                        // Stop
                        clipData.sourceDataLine.stop();
                        //System.out.println("stopepd");
                        clipData.sourceDataLine.close();
                        //System.out.println("closed");
                    }
                }.start();
            }
             */
            
            clips.put((group != null) ? group : name, clipData);
            
            logger.fine("started to play {0}", name);
            
        } catch (Exception e) {
            if (! SoundBank.displayedSoundException)
                logger.warning("couldnt create clip for {0}, {1}", name, e);
            SoundBank.displayedSoundException = true;
        }
    }
    
    public void stop(Collection<String> names) {
        for (String name : names)
            stop(name);
    }
    
    public void stop(String name) {
        if (name == null)
            return;
        
        ClipData clipData = clips.remove(name);
        if (clipData != null)
            clipData.close();
    }
    
    /** updates the sounds related to this source with pan and gain controls */
    public void update(SimpleVector sourcePos, SimpleVector destPos, SimpleVector destDirection) {
        if (clips.isEmpty())
            return;
        
        if (KConfig.dynamicSounds) {
            pan = calculatePan(sourcePos, destPos, destDirection);
            gain = calculateGain(sourcePos, destPos, destDirection);
        }
        
        for (Iterator<ClipData> i = clips.values().iterator(); i.hasNext(); ) {
            ClipData clipData = i.next();
            if (! clipData.isRunning()) {
                clipData.close();
                i.remove();
                continue;
            }
            if (KConfig.dynamicSounds)
                setPanAndGain(clipData, pan, gain);
        }
    }
    /** called when sound settings (volume) changed */
    void update() {
        for (ClipData clipData : clips.values())
            setPanAndGain(clipData, pan, gain);
    }
    
    private void setPanAndGain(ClipData clipData, float pan, float gain) {
        gain += SoundBank.overAllGain + clipData.soundData.initialGain;
        if (pan != 0)
            gain -= 1f;
        if ((clipData.gain != null) && (clipData.gain.getValue() != gain))
            clipData.gain.setValue(gain);
        
        if ((clipData.pan != null) && (clipData.pan.getValue() != pan))
            clipData.pan.setValue(pan);
    }
    
    private boolean tooFarToPlay(SimpleVector sourcePos, SimpleVector destPos) {
        return sourcePos.calcSub(destPos).length() / KConfig.oneMeter > KConfig.soundDiscardDistance;
    }
    
    public void dispose() {
        closeAll();
        soundBank.pointSounds.remove(this);
    }
    
    public void closeAll() {
        for (ClipData clip : clips.values()) {
            clip.close();
        }
        clips.clear();
    }
    
    private float calculatePan(SimpleVector sourcePos, SimpleVector destPos, SimpleVector destDirection) {
        SimpleVector distance = Util3D.Vector.horizontal(destPos.calcSub(sourcePos));
        
        float distLength = distance.length();
        if (distLength == 0)
            return 0f;
        
        SimpleVector direction = Util3D.Vector.horizontal(destDirection);
        float dirLength = direction.length();
        if (dirLength == 0)
            return 0f;
        
        SimpleVector cross = distance.calcCross(direction);
        float sinAlfa = cross.length() / distLength / dirLength;
        
        return ((cross.y > 0) ^ KConfig.reverseStereo) ? sinAlfa : -sinAlfa;
    }
    
    private float calculateGain(SimpleVector sourcePos, SimpleVector destPos, SimpleVector destDirection) {
        SimpleVector distance = destPos.calcSub(sourcePos);
        float distSquare = distance.x*distance.x + distance.y*distance.y + distance.z*distance.z;
        
        if (distSquare == 0)
            return 0f;
        
        double term = distSquare / KConfig.oneMeter / KConfig.oneMeter / KConfig.halfLoudDistanceSquare;
        //float db = -10 + -10 * (float) Math.log10(term);
        float db = -11 + -10 * (float) Math.log10(term);
        
        return (db < 0) ? db : 0f;
    }
    
    private void checkDisposed() {
        if (disposed)
            throw new IllegalStateException("disposed");
    }
}
