package raft.karga.sound;

import raft.karga.core.*;
import raft.karga.core.Logger;
import raft.karga.world.KConfig;
import com.threed.jpct.*;

import java.util.*;
import java.io.*;
import javax.sound.sampled.*;

/**
 *
 * @author hakan eryargi (r a f t)
 */
public class SoundBank {
    public static final boolean INCLUDE_PARENT = false;
    public static final boolean ONLY_SELF = false;
    
    static final boolean PLAY_LOOP = true;
    static final boolean PLAY_ONCE = false;
    
    private static final int MAX_SAFE_CLIP_LENGTH = 2097151;
    
    private final Logger logger = Logger.getLogger();
    
    private final SoundsInfo soundsInfo;
    private final SoundBank parent;
    private final Map<String, Sound> sounds = new HashMap<String, Sound>();
    
    List<PointSound> pointSounds = new ArrayList<PointSound>();
    List<AmbientSound> ambientSounds = new ArrayList<AmbientSound>();
    
    public static int overAllGain = KConfig.overAllGain;
    
    /** used to display exception at most once */
    public static boolean displayedSoundException = false;
    
    /** Creates a new instance of SoundBank */
    public SoundBank(String baseDir) throws Exception {
        this(baseDir, null, null);
    }
    
    /** Creates a new instance of SoundBank */
    public SoundBank(String baseDir, ClassLoader loader) throws Exception {
        this(baseDir, loader, null);
    }
    
    /** Creates a new instance of SoundBank */
    public SoundBank(String baseDir, ClassLoader loader, SoundBank parent) throws Exception {
        this(baseDir, loader, parent, Util.Resource.<SoundsInfo>loadObjectFromResource(
                baseDir + "/" + SoundsInfo.SOUND_INFO_FILE_NAME));
    }
    /** Creates a new instance of SoundBank */
    public SoundBank(String baseDir, ClassLoader loader, SoundBank parent, SoundsInfo soundsInfo) throws Exception {
        this.parent = parent;
        this.soundsInfo = soundsInfo;
        
        for (SoundsInfo.Sound info : soundsInfo.sounds.values()) {
            try {
                Sound sound = new Sound(info, baseDir, loader);
                sounds.put(info.name, sound);
            } catch (Exception e) {
                logger.warning("couldnt load {0}, {1}", info, e);
            }
        }
    }
    
    Sound getSound(String name) {
        Sound sound = sounds.get(name);
        if (sound != null)
            return sound;
        else if (parent != null)
            return parent.getSound(name);
        else return null;
    }
    
    public PointSound createPointSound() {
        PointSound p = new PointSound(this);
        pointSounds.add(p);
        return p;
    }
    
    public AmbientSound createAmbientSound(String name) throws Exception {
        AmbientSound a = new AmbientSound(this, name);
        ambientSounds.add(a);
        return a;
    }
    
    /** plays ambient sound and disposes it */
    public void playAmbientSound(String name) {
        if (KConfig.mute)
            return;
        
        try {
            createAmbientSound(name).playOnceAndDispose();
        } catch (Exception e) {
            if (! displayedSoundException)
                logger.warning("couldnt create ambient sound {0}, {1}", name, e);
            displayedSoundException = true;
        }
    }
    
    public void dispose() {
        List<PointSound> tmpPointSounds = new ArrayList<PointSound>(pointSounds);
        List<AmbientSound> tmpAmbientSounds = new ArrayList<AmbientSound>(ambientSounds);
        
        ambientSounds.clear();
        pointSounds.clear();
        
        for (PointSound pointSound : tmpPointSounds)
            pointSound.dispose();
        for (AmbientSound ambientSound : tmpAmbientSounds)
            ambientSound.dispose();
    }
    
    /** closes all open lines, point sources arent disposed but ambient sounds are */
    public void closeAll(boolean includeParent) {
        closeAllBut(includeParent, null, null);
        /*
        for (PointSound pointSound : pointSounds)
            pointSound.closeAll();
        
        List<AmbientSound> tmpAmbientSounds = new ArrayList<AmbientSound>(ambientSounds);
        for (AmbientSound ambientSound : tmpAmbientSounds)
            ambientSound.dispose();
        
        if (includeParent && (parent != null))
            parent.closeAll(includeParent);
         */
    }
    
    /** closes all open lines, point sources arent disposed but ambient sounds are */
    public void closeAllBut(boolean includeParent, Collection<PointSound> exceptPointSounds,
            Collection<AmbientSound> exceptAmbientSounds) {
        
        for (PointSound pointSound : pointSounds)
            if ((exceptPointSounds == null) || exceptPointSounds.contains(pointSound))
                pointSound.closeAll();
        
        List<AmbientSound> tmpAmbientSounds = new ArrayList<AmbientSound>(ambientSounds);
        for (AmbientSound ambientSound : tmpAmbientSounds)
            if ((exceptAmbientSounds == null) || exceptAmbientSounds.contains(ambientSound))
                ambientSound.dispose();
        
        if (includeParent && (parent != null))
            parent.closeAllBut(includeParent, exceptPointSounds, exceptAmbientSounds);
    }
    
    public void soundSettingsChanged(boolean includeParent) {
        for (PointSound pointSound : pointSounds)
            pointSound.update();
        for (AmbientSound ambientSound : ambientSounds)
            ambientSound.update();
        
        if (includeParent && (parent != null))
            parent.soundSettingsChanged(includeParent);
    }
    
    class Sound {
        public final DataLine.Info lineInfo;
        public final AudioFormat format;
        public final byte[] bytes;
        public final float initialGain;
        public final boolean useClip;
        
        private Sound(SoundsInfo.Sound soundsInfo, String baseDir, ClassLoader loader) throws Exception {
            this(Util.Resource.getResourceAsStream(baseDir + "/" + soundsInfo.file, loader), soundsInfo.initialGain);
            logger.fine("loaded sound for {0}", soundsInfo);
        }
        
        private Sound(InputStream is) throws Exception {
            this(is, 0f);
        }
        
        private Sound(InputStream is, float initialGain) throws Exception {
            this.initialGain = initialGain;
            
            AudioInputStream baseStream = AudioSystem.getAudioInputStream(is);
            AudioFormat baseFormat = baseStream.getFormat();
            
            if (baseFormat.getFrameRate() == AudioSystem.NOT_SPECIFIED) {
                // compressed sound, uncompress it
                this.format = new AudioFormat(
                        AudioFormat.Encoding.PCM_SIGNED,
                        baseFormat.getSampleRate(),
                        16,
                        baseFormat.getChannels(),
                        baseFormat.getChannels() * 2,
                        baseFormat.getSampleRate(),
                        false);
                
                AudioInputStream convertedStream = AudioSystem.getAudioInputStream(format, baseStream);
                
                this.bytes = Util.Streams.readBytes(convertedStream);
                //this.lineInfo = new DataLine.Info(SourceDataLine.class, format);
                baseStream.close();
                
            } else {
                this.format = baseFormat;
                this.bytes = Util.Streams.readBytes(baseStream);
                //this.lineInfo = new DataLine.Info(SourceDataLine.class, format);
            }
            
            useClip = bytes.length < MAX_SAFE_CLIP_LENGTH;
            lineInfo = new DataLine.Info(useClip ? Clip.class : SourceDataLine.class, format);
            
            if (! AudioSystem.isLineSupported(lineInfo))
                throw new UnsupportedAudioFileException(String.valueOf(lineInfo));
        }
    }
}
